from glob import glob
from multiprocessing import Pool, cpu_count
import os
from typing import Dict, List
from tqdm.auto import tqdm
import numpy as np
import pandas as pd
from tabulate import tabulate
import csv
import gzip
import warnings
from .extract import load_epigenomes_table, load_accession_path


def compute_header(statistics: Dict[str, bool]) -> str:
    """Return the header for extracted epigenomes.

    Parameters
    ------------------
    statistics:Dict[str, bool]:
        The statistics to be computed

    Returns
    ------------------
    The header separated with tabs.
    """
    return "\t".join([
        "chrom", "chromStart", "chromEnd", "strand",
        *[s for s, enabled in statistics.items() if enabled]
    ])+'\n'


def get_callback(statistic: str):
    return {
        "mean": np.nanmean,
        "var": np.nanvar,
        "max": np.nanmax,
        "min": np.nanmin,
        "median": np.nanmedian
    }[statistic]


def get_target_path(root: str, cell_line: str, assembly: str, assay_term_name: str, target: str) -> str:
    """Return path where the target epigenomic data are to be stored.

    Parameters
    -----------------------
    root: str,
        Root from where to search the files
        to concatenate.
    cell_line: str,
        Cell line to consider.
    assembly: str,
        The genomic assembly of the data to be retrieved.
    assay_term_name: str,
        Name of the experiment.
    target: str,
        The name of the genomic target.

    Returns
    -----------------------
    The path where to store the epigenomic data.
    """
    return "{root}/{cell_line}/{file_name}.csv.gz".format(
        root=root,
        cell_line=cell_line,
        assembly=assembly,
        file_name=assay_term_name if target == "Unknown" else target
    )


def parse_extracted_epigenome(sources: List[str], target: str, statistics: Dict[str, bool]):
    """Parse the given source bed-like file.

    Parameters
    ----------------------------
    sources: List[str],
        Paths from where to load the sources.
    target: str,
        Epigenomic data target.
    statistics: Dict[str, bool]
        Statistics to be extracted.
    """
    header = compute_header(statistics)
    callbacks = [
        get_callback(s)
        for s, enabled in statistics.items()
        if enabled
    ]
    os.makedirs(os.path.dirname(target), exist_ok=True)

    source_files = [
        gzip.open(source, "rt")
        for source in sources
    ]

    readers = [
        csv.reader(source_file, delimiter='\t')
        for source_file in source_files
    ]
    try:
        with gzip.open(target, "wt") as t:
            # Starting by writing the head
            t.write(header)
            # And now we parse the lines one by one
            for rows in zip(*readers):
                # We extract the values
                chrom, chromStart, chromEnd, _, _, strand = rows[0][:6]
                # Convert the scores to float values
                scores = [
                    [
                        float(s) if s != "NA" else np.nan
                        for s in row[7:]
                    ]
                    for row in rows
                ]
                # Compute the averages in a fully defined way
                averaged_scores = [
                    np.nan
                    if len(sub_scores) == 0 or np.all(np.isnan(sub_scores))
                    else np.nanmean(sub_scores)
                    for sub_scores in scores
                ]
                # Compute the metrics
                metrics = [
                    str(np.nan)
                    if len(averaged_scores) == 0 or np.all(np.isnan(averaged_scores))
                    else cal(averaged_scores).astype(str)
                    for cal in callbacks
                ]
                # And write the results
                t.write("\t".join([
                    chrom, chromStart, chromEnd, strand,
                    *metrics
                ])+'\n')
    except EOFError:
        warnings.warn((
            "Unable to properly finish reading corrupted compressed files {}. "
            "I am now deleting these files. "
            "Just rerun the pipeline to retrieve them again."
        ).format(
            ", ".join(sources)
        ))
        for source in sources:
            os.remove(source)

    for source_file in source_files:
        source_file.close()


def _parse_extracted_epigenome(kwargs: Dict):
    return parse_extracted_epigenome(**kwargs)


def mine(
    root: str,
    statistics: Dict[str, bool],
    cell_lines: List[str],
    assembly: str
):
    """Extract and saves requested statistics from epigenomic files.

    Parameters
    -----------------------
    root: str,
        Root from where to search the files
        to concatenate.
    statistics: Dict[str, bool],
        Dictionary of statistics to extract from windows.
    cell_line: str,
        Cell line to consider.
    assembly: str,
        The genomic assembly of the data to be retrieved.
    """
    tasks = [
        {
            "sources": [
                load_accession_path(root, accession)
                for accession in group.accession
            ],
            "target": get_target_path(root, cell_line, assembly, assay_term_name, target),
            "statistics": statistics
        }
        for (cell_line, assay_term_name, target), group in load_epigenomes_table(cell_lines, assembly).groupby(["cell_line", "assay_term_name", "target"])
        if not os.path.exists(get_target_path(root, assembly, cell_line, assay_term_name, target))
    ]

    with Pool(cpu_count()) as p:
        list(tqdm(
            p.imap(
                _parse_extracted_epigenome,
                tasks
            ),
            desc="Parse extracted epigenomes",
            total=len(tasks)
        ))
        p.close()
        p.join()



import sys
import logging

# Create the logger
logger = logging.getLogger(__name__)
# Change the levels names to that they are 4 chars long
logging.addLevelName(logging.DEBUG, 'DEBG')
logging.addLevelName(logging.WARNING, 'WARN')
logging.addLevelName(logging.ERROR, 'ERRO')
logging.addLevelName(logging.CRITICAL, 'CRIT')
# Set the default log level
logger.setLevel(logging.INFO)
# Set the format of the loger
formatter = logging.Formatter("[%(levelname)s] %(asctime)-15s : %(message)s")

# Setup a stdout logger
shandler = logging.StreamHandler(sys.stdout)
shandler.setLevel(logging.INFO)
shandler.setFormatter(formatter)
logger.addHandler(shandler)



"""Submodule relative to cell lines utilities."""
from typing import List


def get_cell_lines() -> List[str]:
    """Return list of the available cell lines."""
    return [
        "A549",
        "GM12878",
        "H1",
        "HEK293",
        "HepG2",
        "K562",
        "MCF-7"
    ]



import pandas as pd
import os
from tqdm.auto import tqdm
from typing import List
from glob import glob
from multiprocessing import Pool, cpu_count
from .extract import load_epigenomes_table
from .mine import get_target_path


def to_dict(path: str):
    feature_name = path.split(os.sep)[-1].split(".")[0]
    df = pd.read_csv(path, sep="\t", index_col=[0, 1, 2, 3]).round(2)
    df.columns = pd.MultiIndex.from_product([[feature_name], df.columns])
    return df


def concatenate(
    root: str,
    cell_lines: List[str],
    workers: int
):
    """Concatenate the targets into a single file.

    Parameters
    ----------------------------------
    root: str,
        Root from where to search the files
        to concatenate.
    cell_lines: List[str],
        Cell lines to consider.
    workers: int,
        Workers to use to parallelize the loading.
    """
    if workers == -1:
        workers = cpu_count()
    with Pool(min(cpu_count(), workers)) as p:
        for cell_line in tqdm([
            directory_name
            for directory_name in os.listdir(root)
            if os.path.isdir(f"{root}/{directory_name}")
        ], leave=False, desc="Concatenating cell lines"):
            path = "{root}/{cell_line}.csv.xz".format(
                root=root,
                cell_line=cell_line
            )
            paths = glob("{root}/{cell_line}/*.csv.gz".format(
                root=root,
                cell_line=cell_line
            ))

            if os.path.exists(path):
                continue

            pd.concat(list(tqdm(
                p.imap(to_dict, paths),
                desc="Concatenating files",
                total=len(paths),
                leave=False
            )), axis=1).reset_index().to_csv(path, index=False)

        p.close()
        p.join()



"""Generic utilities used in the package."""
from .normalize_epigenomic_data import normalize_epigenomic_data
from .cell_lines import get_cell_lines

__all__ = [
    "normalize_epigenomic_data",
    "get_cell_lines"
]



import pandas as pd
import os
from typing import List
from encodeproject import download
from pybwtool import extract as extract_bigwig
from tqdm.auto import tqdm
from multiprocessing import Pool, cpu_count
import warnings


def extraction_job(
    bed_path: str,
    epigenome_path: str,
    target_path: str,
    url: str,
    clear_download: bool
):
    # Download file if it does not already exist
    if not os.path.exists(epigenome_path):
        download(url, epigenome_path)

    # Extract the features
    extract_bigwig(
        bed_path=bed_path,
        bigwig_path=epigenome_path,
        target=target_path
    )

    # Remove the bigwig file if required
    if clear_download:
        os.remove(epigenome_path)


def _extraction_job(kwargs):
    return extraction_job(**kwargs)


def load_epigenomes_table(
    cell_lines: List[str],
    assembly: str,
) -> pd.DataFrame:
    """Return epigenomic data table.

    Parameters
    ----------------------------
    cell_lines:List[str],
        List of cell lines whose epigenomes are to retrieve.
    assembly: str,
        The genomic assembly of the data to be retrieved.

    Returns
    ----------------------------
    Pandas DataFrame with epigenomes meta data.
    """
    # Loading the epigenomes metadata
    epigenomes = pd.read_csv(
        "{pwd}/epigenomes_metadata/{assembly}.csv".format(
            pwd=os.path.dirname(os.path.abspath(__file__)),
            assembly=assembly
        )
    )
    # Filtering epigenomes for required cell lines
    return epigenomes[epigenomes.cell_line.isin(cell_lines)]


def load_accession_path(root: str, accession: str) -> str:
    return "{root}/{accession}.bed.gz".format(
        root=root,
        accession=accession
    )


def build_extraction_tasks(
    bed_path: str,
    cell_lines: List[str],
    assembly: str,
    epigenomes_path: str,
    target_path: str,
    clear_download: bool
) -> List:
    """Download bigwigs from ENCODE and extract regions specified in given bed file.

    Parameters
    --------------------------
    bed_path:str,
        Either path to the bed file containing the regions of interest.
    cell_lines:List[str],
        List of cell lines whose epigenomes are to retrieve.
    assembly: str,
        The genomic assembly of the data to be retrieved.
    epigenomes_path:str,
        Path where to store the epigenomic bigWig.
    target_path:str,
        Path where to store the epigenomic bed.
    clear_download: bool = False,
        Whetever to delete the downloaded files or not.
        By default False.

    Returns
    --------------------------
    Returns list of tasks to be executed.
    """
    # Loading the epigenomes metadata
    epigenomes = load_epigenomes_table(cell_lines, assembly)
    # Build the tasks
    return [
        {
            "bed_path": bed_path,
            # Where to store the downloaded bigWig file
            "epigenome_path": "{epigenomes_path}/{accession}.{file_format}".format(
                epigenomes_path=epigenomes_path,
                **epigenome.to_dict()
            ),
            # Where to store the extracted regions
            "target_path": load_accession_path(target_path, epigenome.accession),
            "url": epigenome.url,
            "clear_download": clear_download
        }
        for _, epigenome in epigenomes.iterrows()
        if not os.path.exists("{target_path}/{accession}.bed.gz".format(target_path=target_path, **epigenome.to_dict()))
    ]


def extract(
    bed_path: str,
    cell_lines: List[str],
    assembly: str,
    epigenomes_path: str = "epigenomes",
    targets_path: str = "targets",
    clear_download: bool = False,
    workers: int = -1
):
    """Download bigwigs from ENCODE and extract regions specified in given bed file.

    Parameters
    --------------------------
    bed_path:str,
        Either path to the bed file containing the regions of interest.
    cell_lines:List[str],
        List of cell lines whose epigenomes are to retrieve.
    assembly: str,
        The genomic assembly of the data to be retrieved.
    epigenomes_path:str,
        Path where to store the epigenomic bigWig.
    targets_path:str,
        Path where to store the epigenomic bed.
    clear_download: bool = False,
        Whetever to delete the downloaded files or not.
        By default False.
    workers: int = -1,
        Number of workers to use.
        The default, -1, set the workers to the maximum available.

    Raises
    --------------------------
    ValueError,
        If workers number is neiter -1 nor a strictly positive integer.
    ValueError,
        If given nan threshold is not a float value between 0 and 1.
    """
    # Creating target directory if doesn't exist already
    os.makedirs(epigenomes_path, exist_ok=True)
    os.makedirs(targets_path, exist_ok=True)
    # Create the building tasks list
    tasks = build_extraction_tasks(
        bed_path,
        cell_lines,
        assembly,
        epigenomes_path,
        targets_path,
        clear_download
    )
    # Set workers number
    if workers == -1:
        workers = cpu_count()
    workers = min(workers, cpu_count(), len(tasks))
    if len(tasks) == 0:
        warnings.warn("No epigenome still to be parsed, moving on.",
                      category=UserWarning)
        return
    if workers < 1:
        raise ValueError(
            "Given workers number {} is neither -1 or a strictly positive integer.".format(workers))
    # Downloading and elaborating data
    with Pool(workers) as p:
        list(tqdm(
            p.imap(_extraction_job, tasks),
            total=len(tasks),
            desc="Parsing epigenomes"
        ))
        p.close()
        p.join()



from .extract import extract
from .mine import mine
from .concatenate import concatenate
from typing import List


def build(
    bed_path: str,
    cell_lines: List[str],
    assembly: str,
    epigenomes_path: str = "epigenomes",
    targets_path: str = "targets",
    clear_download: bool = False,
    extraction_workers=-1,
    concatenation_workers=-1,
    mine_max: bool = True,
    mine_min: bool = False,
    mine_mean: bool = False,
    mine_median: bool = False,
    mine_variance: bool = False
):
    """Build the dataset.

    Parameters
    --------------------------
    bed_path:str,
        Either path to the bed file containing the regions of interest.
    cell_lines:List[str],
        List of cell lines whose epigenomes are to retrieve.
    assembly: str,
        The genomic assembly of the data to be retrieved.
    epigenomes_path:str,
        Path where to store the epigenomic bigWig.
    targets_path:str,
        Path where to store the epigenomic bed.
    clear_download: bool = False,
        Whetever to delete the downloaded files or not.
        By default False.
    extraction_workers: int = -1,
        Number of workers to use.
        The default, -1, set the workers to the maximum available.
    concatenation_workers: int,
        Workers to use to parallelize the loading.
    mine_max: bool = True,
        Wether to mine max value in windows.
    mine_min: bool = False,
        Wether to mine min value in windows.
    mine_mean: bool = False,
        Wether to mine mean value in windows.
    mine_median: bool = False,
        Wether to mine median value in windows.
    mine_variance: bool = False
        Wether to mine variance value in windows.
    """
    extract(
        bed_path,
        cell_lines,
        assembly,
        epigenomes_path=epigenomes_path,
        targets_path=targets_path,
        clear_download=clear_download,
        workers=extraction_workers
    )
    mine(
        targets_path,
        {
            "max": mine_max,
            "min": mine_min,
            "mean": mine_mean,
            "median": mine_median,
            "var": mine_variance,
        },
        cell_lines,
        assembly
    )
    concatenate(
        targets_path,
        cell_lines,
        concatenation_workers,
    )



"""Module offering methods to retrieve data and tasks for CRR predictions."""
from .build import build
from .load_epigenomes import load_epigenomes
from .load_tasks import (
    active_enhancers_vs_active_promoters,
    active_enhancers_vs_inactive_enhancers,
    inactive_enhancers_vs_inactive_promoters,
    active_promoters_vs_inactive_promoters,
    active_vs_inactive,
    load_all_tasks
)
from .logging import logger

__all__ = [
    "build", "load_epigenomes", "logger",
    "active_enhancers_vs_active_promoters",
    "active_enhancers_vs_inactive_enhancers",
    "inactive_enhancers_vs_inactive_promoters",
    "active_promoters_vs_inactive_promoters",
    "active_vs_inactive",
    "load_all_tasks"
]



"""Module providing normalization for epigenomic data."""
from typing import Tuple
import numpy as np
from sklearn.impute import KNNImputer
from sklearn.preprocessing import StandardScaler, MinMaxScaler


def normalize_epigenomic_data(
    train_x: np.ndarray,
    test_x: np.ndarray = None
) -> Tuple[np.ndarray]:
    """Return imputed and normalized epigenomic data.

    We fit the imputation and normalization on the training data and
    apply it to both the training data and the test data.

    Parameters
    -------------------------
    train_x: np.ndarray,
        Training data to use to fit the imputer and scaled.
    test_x: np.ndarray = None,
        Test data to be normalized.

    Returns
    -------------------------
    Tuple with imputed and scaled train and test data.
    """
    # Create the imputer and scaler object
    imputer = KNNImputer()
    scaler = MinMaxScaler()
    # Fit the imputer object
    imputer.fit(train_x)
    # Impute the train and test data
    imputed_train_x = imputer.transform(train_x)
    if test_x is not None:
        imputed_test_x = imputer.transform(test_x)
    # Fit the scaler object
    scaler.fit(imputed_train_x)
    # Scale the train and test data
    scaled_train_x = scaler.transform(imputed_train_x)
    if test_x is not None:
        scaled_test_x = scaler.transform(imputed_test_x)
    if test_x is not None:
        # Return the normalized data
        return scaled_train_x, scaled_test_x
    return scaled_train_x


"""Module providing straightforward methods to load the tasks."""
from typing import Tuple, Dict
import pandas as pd
from tqdm.auto import tqdm
from .load_epigenomes import load_epigenomes


def load_task(
    cell_line: str = "K562",
    assembly: str = "hg38",
    dataset: str = "fantom",
    metric: str = "mean",
    window_size: int = 256,
    root: str = "datasets",
    verbose: int = 2,
    only_active: bool = False,
    only_inactive: bool = False,
    load_promoters: bool = False,
    load_enhancers: bool = False,
    binarize: bool = False,
    min_active_tpm_value: float = 1,
    max_inactive_tpm_value: float = 1,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Return epigenomic data and labels for given parameters.

    Parameters
    ----------------------------------------
    cell_line: str = "K562",
        Cell line to consider. By default K562.
        Currently available cell lines are
        listed in the repository README file.
    assembly: str,
        The genomic assembly of the data to be retrieved.
    dataset: str = "fantom",
        Dataset to consider. By default fantom.
        Currently available datasets are
        listed in the repository README file.
    metric: str = "mean",
        The metric to load.
    window_size: int = 256,
        Window size to consider. By default 256.
        Currently available window sizes are
        listed in the repository README file.
    root: str = "datasets"
        Where to store the downloaded data.
    verbose: int = 2,
        Verbosity level.
    only_active: bool = False,
        Wether to filter for only active.
    only_inactive: bool = False,
        Wether to filter for only inactive.
    load_promoters: bool = False,
        Wether to load promoters.
    load_enhancers: bool = False
        Wether to load enhancers.
    binarize: bool = False,
        Whether to binarize the TPM values.
    min_active_tpm_value: float = 1,
        Minimum TPM value.
    max_inactive_tpm_value: float = 1,
        Maximum TPM value.
        Values between the minimum and maximum will be dropped.

    Returns
    ----------------------------------------
    Return tuple with input and output DataFrames.
    """
    if only_active and only_inactive:
        raise ValueError(
            "It is not possible to require both only active and inactive."
        )
    if not (load_promoters or load_enhancers):
        raise ValueError(
            "You need to load either the promoters or the enhancers."
        )
    if (only_active or only_inactive) and not (load_promoters and load_enhancers):
        raise ValueError(
            "When requiring to filter only active or inactive regions, "
            "you must load both enhancers and promoters."
        )

    (promoters_epi, promoters_labels), (enhancers_epi, enhancers_labels) = [
        load_epigenomes(
            cell_line=cell_line,
            assembly=assembly,
            dataset=dataset,
            region=region,
            metric=metric,
            window_size=window_size,
            root=root,
            binarize=binarize,
            min_active_tpm_value=min_active_tpm_value,
            max_inactive_tpm_value=max_inactive_tpm_value,
            verbose=verbose
        ) if enabled else (None, None)
        for region, enabled in (
            ("promoters", load_promoters),
            ("enhancers", load_enhancers),
        )
    ]
    if only_active:
        promoters_epi = promoters_epi[promoters_labels.to_numpy() == 1]
        enhancers_epi = enhancers_epi[enhancers_labels.to_numpy() == 1]
        promoters_labels = promoters_labels[promoters_labels.to_numpy() == 1]
        enhancers_labels = enhancers_labels[enhancers_labels.to_numpy() == 1]
        enhancers_labels[enhancers_labels.columns[0]] = 0
    elif only_inactive:
        promoters_epi = promoters_epi[promoters_labels.to_numpy() == 0]
        enhancers_epi = enhancers_epi[enhancers_labels.to_numpy() == 0]
        promoters_labels = promoters_labels[promoters_labels.to_numpy() == 0]
        promoters_labels[promoters_labels.columns[0]] = 1
        enhancers_labels = enhancers_labels[enhancers_labels.to_numpy() == 0]

    if only_active or only_inactive:
        epigenomic_data = pd.concat([
            promoters_epi,
            enhancers_epi
        ])
        labels = pd.concat([
            promoters_labels,
            enhancers_labels
        ])
    else:
        epigenomic_data = pd.concat([
            region
            for region in (promoters_epi, enhancers_epi)
            if region is not None
        ])
        labels = pd.concat([
            region
            for region in (promoters_labels, enhancers_labels)
            if region is not None
        ])

    return epigenomic_data, labels


def active_promoters_vs_inactive_promoters(
    cell_line: str = "K562",
    assembly: str = "hg38",
    dataset: str = "fantom",
    metric: str = "mean",
    window_size: int = 256,
    root: str = "datasets",
    binarize: bool = False,
    min_active_tpm_value: float = 1,
    max_inactive_tpm_value: float = 1,
    verbose: int = 2,
):
    """Return epigenomic data and labels for given parameters.

    Parameters
    ----------------------------------------
    cell_line: str = "K562",
        Cell line to consider. By default K562.
        Currently available cell lines are
        listed in the repository README file.
    assembly: str,
        The genomic assembly of the data to be retrieved.
    dataset: str = "fantom",
        Dataset to consider. By default fantom.
        Currently available datasets are
        listed in the repository README file.
    metric: str = "mean",
        The metric to load.
    window_size: int = 256,
        Window size to consider. By default 256.
        Currently available window sizes are
        listed in the repository README file.
    root: str = "datasets"
        Where to store the downloaded data.
    binarize: bool = False,
        Whether to binarize the TPM values.
    min_active_tpm_value: float = 1,
        Minimum TPM value.
    max_inactive_tpm_value: float = 1,
        Maximum TPM value.
        Values between the minimum and maximum will be dropped.
    verbose: int = 2,
        Verbosity level.

    Returns
    ----------------------------------------
    Return tuple with input and output DataFrames.
    """
    return load_task(
        cell_line=cell_line,
        assembly=assembly,
        dataset=dataset,
        metric=metric,
        window_size=window_size,
        root=root,
        verbose=verbose,
        only_active=False,
        only_inactive=False,
        load_enhancers=False,
        load_promoters=True,
        binarize=binarize,
        min_active_tpm_value=min_active_tpm_value,
        max_inactive_tpm_value=max_inactive_tpm_value,
    )


def active_enhancers_vs_inactive_enhancers(
    cell_line: str = "K562",
    assembly: str = "hg38",
    dataset: str = "fantom",
    metric: str = "mean",
    window_size: int = 256,
    root: str = "datasets",
    binarize: bool = False,
    min_active_tpm_value: float = 1,
    max_inactive_tpm_value: float = 1,
    verbose: int = 2,
):
    """Return epigenomic data and labels for given parameters.

    Parameters
    ----------------------------------------
    cell_line: str = "K562",
        Cell line to consider. By default K562.
        Currently available cell lines are
        listed in the repository README file.
    assembly: str,
        The genomic assembly of the data to be retrieved.
    dataset: str = "fantom",
        Dataset to consider. By default fantom.
        Currently available datasets are
        listed in the repository README file.
    metric: str = "mean",
        The metric to load.
    window_size: int = 256,
        Window size to consider. By default 256.
        Currently available window sizes are
        listed in the repository README file.
    root: str = "datasets"
        Where to store the downloaded data.
    binarize: bool = False,
        Whether to binarize the TPM values.
    min_active_tpm_value: float = 1,
        Minimum TPM value.
    max_inactive_tpm_value: float = 1,
        Maximum TPM value.
        Values between the minimum and maximum will be dropped.
    verbose: int = 2,
        Verbosity level.

    Returns
    ----------------------------------------
    Return tuple with input and output DataFrames.
    """
    return load_task(
        cell_line=cell_line,
        assembly=assembly,
        dataset=dataset,
        metric=metric,
        window_size=window_size,
        root=root,
        verbose=verbose,
        only_active=False,
        only_inactive=False,
        load_enhancers=True,
        load_promoters=False,
        binarize=binarize,
        min_active_tpm_value=min_active_tpm_value,
        max_inactive_tpm_value=max_inactive_tpm_value,
    )


def active_enhancers_vs_active_promoters(
    cell_line: str = "K562",
    assembly: str = "hg38",
    dataset: str = "fantom",
    metric: str = "mean",
    window_size: int = 256,
    root: str = "datasets",
    verbose: int = 2,
    **kwargs: Dict
):
    """Return epigenomic data and labels for given parameters.

    Parameters
    ----------------------------------------
    cell_line: str = "K562",
        Cell line to consider. By default K562.
        Currently available cell lines are
        listed in the repository README file.
    assembly: str,
        The genomic assembly of the data to be retrieved.
    dataset: str = "fantom",
        Dataset to consider. By default fantom.
        Currently available datasets are
        listed in the repository README file.
    metric: str = "mean",
        The metric to load.
    window_size: int = 256,
        Window size to consider. By default 256.
        Currently available window sizes are
        listed in the repository README file.
    root: str = "datasets"
        Where to store the downloaded data.
    verbose: int = 2,
        Verbosity level.

    Returns
    ----------------------------------------
    Return tuple with input and output DataFrames.
    """
    return load_task(
        cell_line=cell_line,
        assembly=assembly,
        dataset=dataset,
        metric=metric,
        window_size=window_size,
        root=root,
        verbose=verbose,
        only_active=True,
        only_inactive=False,
        load_enhancers=True,
        load_promoters=True,
    )


def inactive_enhancers_vs_inactive_promoters(
    cell_line: str = "K562",
    assembly: str = "hg38",
    dataset: str = "fantom",
    metric: str = "mean",
    window_size: int = 256,
    root: str = "datasets",
    verbose: int = 2,
    **kwargs: Dict
):
    """Return epigenomic data and labels for given parameters.

    Parameters
    ----------------------------------------
    cell_line: str = "K562",
        Cell line to consider. By default K562.
        Currently available cell lines are
        listed in the repository README file.
    assembly: str,
        The genomic assembly of the data to be retrieved.
    dataset: str = "fantom",
        Dataset to consider. By default fantom.
        Currently available datasets are
        listed in the repository README file.
    metric: str = "mean",
        The metric to load.
    window_size: int = 256,
        Window size to consider. By default 256.
        Currently available window sizes are
        listed in the repository README file.
    root: str = "datasets"
        Where to store the downloaded data.
    verbose: int = 2,
        Verbosity level.

    Returns
    ----------------------------------------
    Return tuple with input and output DataFrames.
    """
    return load_task(
        cell_line=cell_line,
        assembly=assembly,
        dataset=dataset,
        metric=metric,
        window_size=window_size,
        root=root,
        verbose=verbose,
        only_active=False,
        only_inactive=True,
        load_enhancers=True,
        load_promoters=True,
    )


def load_all_tasks(
    cell_line: str = "K562",
    assembly: str = "hg38",
    dataset: str = "fantom",
    metric: str = "mean",
    window_size: int = 256,
    root: str = "datasets",
    verbose: int = 2,
    leave: bool = False,
    binarize: bool = False,
    min_active_tpm_value: float = 1,
    max_inactive_tpm_value: float = 1,
):
    """Return generator with all the tasks.

    Parameters
    ----------------------------------------
    cell_line: str = "K562",
        Cell line to consider. By default K562.
        Currently available cell lines are
        listed in the repository README file.
    assembly: str,
        The genomic assembly of the data to be retrieved.
    dataset: str = "fantom",
        Dataset to consider. By default fantom.
        Currently available datasets are
        listed in the repository README file.
    metric: str = "mean",
        The metric to load.
    window_size: int = 256,
        Window size to consider. By default 256.
        Currently available window sizes are
        listed in the repository README file.
    root: str = "datasets"
        Where to store the downloaded data.
    verbose: int = 2,
        Verbosity level.
    leave: bool = False,
        Wether to leave the loading bar.
    binarize: bool = False,
        Whether to binarize the TPM values.
    min_active_tpm_value: float = 1,
        Minimum TPM value.
    max_inactive_tpm_value: float = 1,
        Maximum TPM value.
        Values between the minimum and maximum will be dropped.

    Returns
    ----------------------------------------
    Return tuple with input and output DataFrames.
    """
    return (
        (
            task(
                cell_line=cell_line,
                assembly=assembly,
                dataset=dataset,
                metric=metric,
                window_size=window_size,
                root=root,
                verbose=verbose,
                binarize=binarize,
                min_active_tpm_value=min_active_tpm_value,
                max_inactive_tpm_value=max_inactive_tpm_value,
            ),
            task.__name__
        )
        for task in tqdm(
            (
                active_enhancers_vs_inactive_enhancers,
                active_promoters_vs_inactive_promoters,
                active_enhancers_vs_active_promoters,
                inactive_enhancers_vs_inactive_promoters,
            ),
            desc="Executing CRR prediction tasks",
            disable=verbose == 0,
            leave=leave
        )
    )



from typing import Tuple
import os
from downloaders import BaseDownloader
import pandas as pd


def load_epigenomes(
    cell_line: str = "K562",
    assembly: str = "hg38",
    dataset: str = "fantom",
    region: str = "promoters",
    metric: str = "mean",
    window_size: int = 256,
    root: str = "datasets",
    binarize: bool = False,
    min_active_tpm_value: float = 1,
    max_inactive_tpm_value: float = 1,
    verbose: int = 2
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Return epigenomic data and labels for given parameters.

    Parameters
    ----------------------------------------
    cell_line: str = "K562",
        Cell line to consider. By default K562.
        Currently available cell lines are
        listed in the repository README file.
    assembly: str,
        The genomic assembly of the data to be retrieved.
    dataset: str = "fantom",
        Dataset to consider. By default fantom.
        Currently available datasets are
        listed in the repository README file.
    region: str = "promoters",
        Region to consider. By default promoters.
        Currently available region are
        listed in the repository README file.
    metric: str = "mean",
        The metric to load.
    window_size: int = 200,
        Window size to consider. By default 200.
        Currently available window sizes are
        listed in the repository README file.
    root: str = "datasets"
        Where to store the downloaded data.
    binarize: bool = False,
        Whether to binarize the TPM values.
    min_active_tpm_value: float = 1,
        Minimum TPM value.
    max_inactive_tpm_value: float = 1,
        Maximum TPM value.
        Values between the minimum and maximum will be dropped.
    verbose: int = 2,
        Verbosity level.

    Returns
    ----------------------------------------
    Return tuple with input and output DataFrames.
    """

    if max_inactive_tpm_value > min_active_tpm_value:
        raise ValueError(
            "The maximum inactive TPM value is higher ",
            "than the given minimum active TPM value."
        )

    repository = "https://github.com/LucaCappelletti94/epigenomic_dataset/blob/master/preprocessed"
    get_parameter = "?raw=true"
    data_path_placeholder = "{{root}}/{dataset}/{assembly}/{window_size}/{region}/{cell_line}.csv.xz".format(
        root=root,
        dataset=dataset,
        assembly=assembly,
        window_size=window_size,
        region=region,
        cell_line=cell_line
    )
    data_path = data_path_placeholder.format(root=root)
    label_path_placeholder = "{{root}}/{dataset}/{assembly}/{window_size}/{region}.bed.xz".format(
        root=root,
        dataset=dataset,
        assembly=assembly,
        window_size=window_size,
        region=region
    )
    label_path = label_path_placeholder.format(root=root)

    downloader = BaseDownloader(target_directory=root, verbose=verbose)

    downloader.download(
        urls=data_path_placeholder.format(root=repository)+get_parameter,
        paths=data_path
    )
    downloader.download(
        urls=label_path_placeholder.format(root=repository)+get_parameter,
        paths=label_path
    )

    dtypes = {
        "chrom": "str",
        "chromStart": "int",
        "chromEnd": "int",
        "strand": "str"
    }

    X = pd.read_csv(
        data_path,
        index_col=[0, 1, 2, 3],
        header=[0, 1],
        low_memory=False,
        dtype=dtypes
    )

    X.index.rename(
        list(dtypes.keys()),
        inplace=True
    )

    X = X[[
        col
        for col in X.columns
        if metric in col
    ]]

    X = X.droplevel(1, axis=1)

    y = pd.read_csv(
        label_path,
        index_col=[0, 1, 2, 3],
        sep="\t",
        low_memory=False,
        dtype=dtypes
    )

    # Making sure the two datasets indices are aligned
    y = y.loc[X.index]

    # If the minimum and maximum values are not equal,
    # we need to drop the unknown values
    if min_active_tpm_value != max_inactive_tpm_value:
        gray_zone_values = (y > min_active_tpm_value) & (y < max_inactive_tpm_value)
        y = y[~gray_zone_values]
        X = X[~gray_zone_values]
    
    if binarize:
        y = y > min_active_tpm_value

    normalized_cell_line = cell_line.replace("-", "").upper()

    if normalized_cell_line not in y.columns:
        raise ValueError(
            (
                "The requested cell line {} is not present in the labels. "
                "The available cell lines are {}"
            ).format(normalized_cell_line, ", ".join(y.columns))
        )

    y = y[[normalized_cell_line]]

    return X, y
